"""Wrapper library providing transparent proxies to configuration objects.

This library provides wrapper classes that act as transparent proxies to ConfigurationDict,
ConfigurationList and Configuration objects. Each wrapper encapsulates the actual configuration
object in a __sconfig__ attribute and forwards method calls to it.
"""

# Development
# pylint: disable=unused-argument, unused-import, too-few-public-methods


# Please create the wrapper.py library. The principle of this library is to provide a transparent proxy to any ConfigurationDict
# or ConfigurationList.

# This Should provide a class override of the ConfigurationList, ConfigurationDict and Configuration, where an instance of those object is encapulsated
# in a __sconfig__ atribute. Then those override class have some dunders methods, directly forwarded to the
# __sconfig__ methods.


from superconf.configuration import Configuration, ConfigurationDict, ConfigurationList

# Provide high level API


class ConfigurationWrapper:
    """Base wrapper class providing transparent proxy functionality.
    
    This class implements the core proxy functionality used by all wrapper classes.
    It forwards attribute access and method calls to the wrapped configuration object.
    """

    def __init__(self, config):
        """Initialize wrapper with configuration object.
        
        Args:
            config: The configuration object to wrap
        """
        self.__sconfig__ = config

    def __getattr__(self, name):
        """Forward attribute access to wrapped config object."""
        return getattr(self.__sconfig__, name)

    def __getitem__(self, key):
        """Forward item access to wrapped config object."""
        return self.__sconfig__[key]

    def __setitem__(self, key, value):
        """Forward item assignment to wrapped config object."""
        self.__sconfig__[key] = value

    def __iter__(self):
        """Forward iteration to wrapped config object."""
        return iter(self.__sconfig__)

    def __len__(self):
        """Forward length queries to wrapped config object."""
        return len(self.__sconfig__)

    def __str__(self):
        """Forward string representation to wrapped config object."""
        return str(self.__sconfig__)

    def __repr__(self):
        """Forward detailed string representation to wrapped config object."""
        return f"{self.__class__.__name__}({repr(self.__sconfig__)})"


class ConfigurationDictWrapper(ConfigurationWrapper):
    """Wrapper for ConfigurationDict objects.
    
    Provides a transparent proxy to a ConfigurationDict instance.
    """

    def __init__(self, config=None, **kwargs):
        """Initialize with optional config object or create new one.
        
        Args:
            config: Existing ConfigurationDict instance or None
            **kwargs: Arguments to create new ConfigurationDict if config is None
        """
        if config is None:
            config = ConfigurationDict(**kwargs)
        super().__init__(config)


class ConfigurationListWrapper(ConfigurationWrapper):
    """Wrapper for ConfigurationList objects.
    
    Provides a transparent proxy to a ConfigurationList instance.
    """

    def __init__(self, config=None, **kwargs):
        """Initialize with optional config object or create new one.
        
        Args:
            config: Existing ConfigurationList instance or None
            **kwargs: Arguments to create new ConfigurationList if config is None
        """
        if config is None:
            config = ConfigurationList(**kwargs)
        super().__init__(config)


class ConfigurationWrap(ConfigurationDictWrapper):
    """Wrapper for Configuration objects.
    
    Provides a transparent proxy to a Configuration instance.
    Inherits from ConfigurationDictWrapper since Configuration extends ConfigurationDict.
    """

    def __init__(self, config=None, **kwargs):
        """Initialize with optional config object or create new one.
        
        Args:
            config: Existing Configuration instance or None
            **kwargs: Arguments to create new Configuration if config is None
        """
        if config is None:
            config = Configuration(**kwargs)
        super().__init__(config)



"Common code"

# pylint: disable=too-few-public-methods
import json
import logging
import os

import yaml

log = logging.getLogger(__name__)


class _NoneType:
    """
    A special type that behaves as a replacement for None.
    We have to put a new default value to know if a variable has been set by
    the user explicitly. This is useful for the ``CommandLine`` loader, when
    CLI parsers force you to set a default value, and thus, break the discovery
    chain.
    """

    def repr(self):
        "Return string representation"
        return "<NONE_TYPE>"

    def __str__(self):
        return self.repr()

    def __repr__(self):
        return self.repr()

    def __bool__(self):
        return False


class NotSet(_NoneType):
    "Represent an unset arg"

    def repr(self):
        return "<NOT_SET>"


class UnSetArg(_NoneType):
    "Represent an unset arg"

    def repr(self):
        return "<UNSET_ARG>"


class Failure(_NoneType):
    "Represent a failure"

    def repr(self):
        return "<FAILURE>"


class Default(_NoneType):
    "Represent a default"

    def repr(self):
        return "<DEFAULT>"


NOT_SET = NotSet()
UNSET_ARG = UnSetArg()
FAIL = UnSetArg()
DEFAULT = Default()
assert UNSET_ARG is not NOT_SET
assert NOT_SET is not UNSET_ARG
assert not isinstance(UNSET_ARG, str)
assert not isinstance(NOT_SET, str)
assert (NOT_SET or "INVALID") == "INVALID"

# assert UNSET_ARG is not NOT_SET
# assert UNSET_ARG == NOT_SET


def from_json(string):
    "Transform JSON string to python dict"
    return json.loads(string)


def to_json(obj, nice=True):
    "Transform JSON string to python dict"
    if nice:
        return json.dumps(obj, indent=2)
    return json.dumps(obj)


def from_yaml(string):
    "Transform YAML string to python dict"
    return yaml.safe_load(string)


def to_yaml(obj):
    "Transform obj to YAML"
    return yaml.dump(obj)

    # # Ruamel support
    # options = {}
    # string_stream = StringIO()

    # if isinstance(obj, str):
    #     obj = json.loads(obj)

    # yaml.dump(obj, string_stream, **options)
    # output_str = string_stream.getvalue()
    # string_stream.close()
    # if not headers:
    #     output_str = output_str.split("\n", 2)[2]
    # return output_str


def read_file(file):
    "Read file content"
    with open(file, encoding="utf-8") as _file:
        return "".join(_file.readlines())


def write_file(file, content, create_dirs=True):
    "Write content to file"

    if create_dirs:
        file_folder = os.path.dirname(file)
        if not os.path.exists(file_folder):
            os.makedirs(file_folder)

    with open(file, "w", encoding="utf-8") as _file:
        _file.write(content)



"Parser library"

# pylint: disable=too-few-public-methods

from typing import Iterator, Tuple, Union

STATE_INITIAL = "initial"
STATE_PARSING_KEY = "parsing_key"
STATE_PARSING_VALUE = "parsing_value"
STATE_PARSING_VALUE_ESCAPE = "parsing_value_escape"
STATE_PARSING_COMMENT = "parsing_comment"

COMMENT = "#"
END_OF_LINE = "\n"
ESCAPE_CHAR = "\\"
SPACES = set(" \n")
QUOTES = set("'\"")


class BufferedStreamReader:
    "Undocumented class"
    BUFFER_SIZE = 1024

    def __init__(self, stream):
        self.stream = stream
        self.buffer = stream.read(self.BUFFER_SIZE)
        self.position = 0

    def read_char(self):
        "Read character"

        if self._is_buffer_depleted():
            self._fill_buffer()
            if not self.buffer:
                return None

        char = self.buffer[self.position]
        self.position += 1
        return char

    def _is_buffer_depleted(self):
        return self.position >= len(self.buffer)

    def _fill_buffer(self):
        self.buffer = self.stream.read(self.BUFFER_SIZE)
        self.position = 0


class EnvFileParser:
    "Environment parser"

    def __init__(self, stream):
        self.state = STATE_INITIAL
        self._stream = BufferedStreamReader(stream)

        self._current_key = []
        self._current_value = []
        self._current_quote = ""
        self._key_parsed = False

    def parse_config(self) -> Iterator[Tuple[str, str]]:
        "Parse config"
        key = self._current_key
        value = self._current_value

        while True:
            char = self._stream.read_char()
            if not char:
                if key or value:
                    yield self._return_current_config()
                break

            if self._process_initial(char):
                continue

            if self._process_quotes(char):
                continue

            if self._start_comment(char):
                continue

            parsed_value = self._finish_comment(char)
            if parsed_value:
                if isinstance(parsed_value, tuple):
                    yield parsed_value
                continue

            if self._parse_key(char):
                continue

            parsed_value = self._parse_value(char)
            if parsed_value:
                if isinstance(parsed_value, tuple):
                    yield parsed_value
                continue

            if self._parse_escaped_value(char):
                continue

    def _start_comment(self, char):
        if char == COMMENT and (not self._current_quote or self.state == STATE_INITIAL):
            self.state = STATE_PARSING_COMMENT
            return True

        return False

    def _finish_comment(self, char) -> Union[bool, Tuple[str, str]]:
        if not (self.state == STATE_PARSING_COMMENT and char == END_OF_LINE):
            return False

        if not (self._current_key or self._current_value) or not self._key_parsed:
            # Key started with a comment
            self._reset_state()
            return True

        return self._return_current_config()

    def _process_quotes(self, char):
        if char not in QUOTES:
            return False
        if not self._current_quote:
            self._current_quote = char
            return True

        if self._current_quote == char:
            self._current_quote = ""
            return True

        self._current_value.append(char)
        return True

    def _process_initial(self, char):
        if self.state != STATE_INITIAL:
            return False

        if self._start_comment(char):
            return True

        if char in SPACES:
            return True

        self._current_key.append(char)
        self.state = STATE_PARSING_KEY
        return True

    def _parse_key(self, char):
        if self.state != STATE_PARSING_KEY:
            return False

        if char == "=":
            self.state = STATE_PARSING_VALUE
            self._key_parsed = True
            return True

        if char == END_OF_LINE:
            self._reset_state()
            return True

        self._current_key.append(char)
        return True

    def _parse_value(self, char) -> Union[bool, Tuple[str, str]]:
        if self.state != STATE_PARSING_VALUE:
            return False

        if char == ESCAPE_CHAR:
            self.state = STATE_PARSING_VALUE_ESCAPE
            return True

        if char == " " and not (self._current_value or self._current_quote):
            return True

        if char == END_OF_LINE:
            return self._return_current_config()

        self._current_value.append(char)
        return True

    def _parse_escaped_value(self, char):
        if self.state != STATE_PARSING_VALUE_ESCAPE:
            return False

        if char == END_OF_LINE:
            # resume value parsing on the next line
            self.state = STATE_PARSING_VALUE
            return True

        # That was a literal \
        self.state = STATE_PARSING_VALUE
        self._current_value.extend([ESCAPE_CHAR, char])
        return True

    def _return_current_config(self) -> Tuple[str, str]:
        key, value = "".join(self._current_key), "".join(self._current_value)
        self._reset_state()
        return key.rstrip(), value.rstrip()

    def _reset_state(self):
        self.state = STATE_INITIAL
        self._current_key.clear()
        self._current_value.clear()
        self._key_parsed = False



"Superconf Exceptions"


class ConfigurationException(Exception):
    "General error"


class InvalidConfigurationFile(ConfigurationException):
    "Invalid configuration File"


class MissingSettingsSection(InvalidConfigurationFile):
    "Missing section setting"


class InvalidPath(ConfigurationException):
    "Configuration Exception"


class InvalidCastConfiguration(ConfigurationException):
    "Raised when an invalid cast configuration is found"


# class CastValueFailure(ConfigurationException, ValueError):
class CastValueFailure(ConfigurationException):
    "Raised when a value can't be casted"


class UndeclaredField(ConfigurationException):
    "Raised when querrying unexisting field"


class UnknownExtraField(ConfigurationException):
    "Raised when trying to set value of undefined field. Enable extra_fields=True to disable"


class UnknownSetting(ConfigurationException):
    "Raised when an unexpected field is met. Enable extra_fields=True to disable this error"



"Fields management"

# pylint: disable=protected-access

# TOFIX:
# pylint: disable=invalid-name


# from pprint import pprint
import logging
from types import SimpleNamespace
from typing import Callable

from superconf import exceptions

from .casts import (
    AsBoolean,
    AsDict,
    AsIdentity,
    AsInt,
    AsList,
    AsOption,
    AsString,
    AsTuple,
)
from .common import FAIL, NOT_SET
from .loaders import _Value

# Shortcuts for standard casts
as_string = AsString()
as_boolean = AsBoolean()
as_int = AsInt()
as_list = AsList()
as_dict = AsDict()
as_tuple = AsTuple()
as_option = AsOption
as_is = AsIdentity()

logger = logging.getLogger(__name__)


class Field:
    """Base class for configuration fields.

    A Field represents a single configuration value with optional type casting,
    default values, and help text. Fields are used as descriptors in configuration
    classes to define the structure and behavior of configuration values.

    Attributes:
        cast: The casting function to use for converting raw values. If None,
            will be determined based on the default value's type.
    """

    cast = None

    # pylint: disable=redefined-builtin
    def __init__(
        self,
        key: str = None,
        *,
        help: str = "",
        default=NOT_SET,
        cast: Callable = None,
    ):
        """Initialize a new Field instance.

        Args:
            key: Name of the value used in file or environment variable.
                Set automatically by the metaclass.
            help: Plain-text description of the value.
            default: Default value if none is provided. If left unset,
                loading a config that fails to provide this value
                will raise a UnknownConfiguration exception.
            cast: Callable to cast variable with. Defaults to type of
                default (if provided), identity if default is not
                provided or raises TypeError if provided cast is not
                callable.
        """
        self.key = key
        self.help = help
        self.default = default
        self.cast = cast or self.cast

    def __get__(self, conf_instance, owner):
        """Descriptor get method to retrieve the field's value.

        Args:
            conf_instance: The configuration instance this field belongs to.
            owner: The class that owns this descriptor.

        Returns:
            The field's value if accessed through an instance,
            or the field itself if accessed through the class.
        """
        if conf_instance:
            return conf_instance.get_field_value(key=self.key, field=self)
        return self

    def __repr__(self):
        """Return a string representation of the field.

        Returns:
            A string showing the field's class name, key, and help text.
        """
        return f'{self.__class__.__name__}(key="{self.key}", help="{self.help}")'

    def is_container(self):
        """Check if this field is a container type.

        Returns:
            bool: True if this field has a children_class attribute,
            indicating it can contain nested configuration values.
        """
        children_class = getattr(self, "children_class", None)
        if children_class is not None:
            return True
        return False

    # pylint: disable=too-many-locals, too-many-branches, too-many-arguments, too-many-statements, too-many-positional-arguments
    def resolve_value(
        self,
        conf_instance,
        value=NOT_SET,
        default=NOT_SET,
        cast=NOT_SET,
        loaders=NOT_SET,
        lookup=True,
        **kwargs,
    ):
        """Resolve the final value for this field.

        This method handles the complex logic of determining the field's value by:
        1. Checking for explicitly provided values
        2. Looking up values through loaders
        3. Falling back to defaults
        4. Applying type casting

        Args:
            conf_instance: The configuration instance this field belongs to.
            value: Explicitly provided value, takes precedence if set.
            default: Override for the field's default value.
            cast: Override for the field's cast function.
            loaders: List of loader objects to use for value lookup.
            lookup: Enable loaders, or only use _value instead.
            **kwargs: Additional keyword arguments passed to loaders.

        Returns:
            tuple: A tuple containing:
                - The resolved and cast value
                - A SimpleNamespace containing metadata about the resolution process

        Raises:
            CastValueFailure: If strict casting is enabled and the value
                cannot be cast to the desired type.
        """
        key = self.key
        assert isinstance(key, (str, int)), f"Got: {type(key)} {key}"

        # Process defaults
        default_from = ["args"]
        if default is NOT_SET and isinstance(conf_instance._default, dict):
            # Fetch default from container

            # default2 = default
            try:
                default = conf_instance.query_inst_cfg(
                    "default", override=kwargs, default=NOT_SET
                )[key]
                default_from.append("conf_instance_query")
            except KeyError:
                pass

        if default is NOT_SET:
            # Fetch default from field
            default = self.default
            default_from.append("field_instance")

        # Process value
        if value is NOT_SET:
            # Fetch default from container
            try:
                value = conf_instance._value[key]
            except (TypeError, KeyError):  # For dict
                pass
            except IndexError:  # For list
                pass

        # Process cast
        cast_from = []
        if cast is NOT_SET:
            cast = self.cast
            cast_from.append(f"field_attr:{self}.cast")
        if cast is NOT_SET:
            cast = conf_instance._cast
            cast_from.append(f"conf_attr:{conf_instance}._cast")

        # Process loaders
        if not lookup:
            loaders = [_Value({key: value})]
        else:
            if loaders is NOT_SET:
                loaders = conf_instance._loaders
            if value:
                # pylint: disable=using-constant-test
                if False:
                    loaders.insert(0, _Value({key: value}))
                else:
                    # Remove any existing _Value
                    loaders = [
                        loader for loader in loaders if not isinstance(loader, _Value)
                    ]
                    loaders.append(_Value({key: value}))

        # Determine cast method
        if callable(cast):
            # cast = cast
            cast_from.append("cast_is_callable")
        elif cast is None and (default is NOT_SET or default is None):
            cast = as_is
            cast_from.append("cast_is_none_and_no_defaults")
        elif isinstance(default, bool):
            cast = as_boolean
            cast_from.append("cast_as_boolean")
        elif cast is None:
            cast = type(default)
            cast_from.append("cast_is_none")
        elif cast is NOT_SET:
            if default is NOT_SET or default is None:
                cast_from.append("cast_notset_type_default")
                cast = type(default)
            else:
                cast_from.append("cast_notset_type_as_is")
                cast = as_is
        else:
            raise TypeError(f"Cast must be callable, got: {type(cast)}")

        # Process things
        is_casted = False
        result = NOT_SET
        loader_from = []
        results_from = []
        # print(f"ACTUAL LOADERS FOR {self}", loaders)
        logger.debug("Current loaders for %s: %s", self, loaders)
        for loader in loaders:
            loader_from.append(str(loader))
            try:
                logger.debug("Try loader for %s.%s: %s", self, key, loaders)
                # print(f"  > LOADER: try search in {loader} key: {key}")
                # print("VS", self, conf_instance)
                result = loader.getitem(conf_instance, key, **kwargs)

            except KeyError:
                continue

            if result is not NOT_SET:
                results_from.append(f"from_loader:{loader}")
                result = cast(result)
                is_casted = True
                break

        # Nothing found in all loaders, then fallback on default
        if result is NOT_SET:
            result = value
            results_from.append("from_value")
        if result is NOT_SET:
            result = default
            results_from.append("from_default")

        # Try to cast value
        if not is_casted:
            error = None
            try:
                result = cast(result)
                results_from.append(f"casted:{cast}")
            except (exceptions.InvalidCastConfiguration, ValueError, TypeError) as err:
                error = err
                result = NOT_SET
                results_from.append(f"not_castable:{cast}")

            # Check for strict_cast mode:
            if error is not None and conf_instance._strict_cast is True:
                msg = (
                    f"Got error {conf_instance}.{key} {type(error)}: {error}, "
                    "set strict_cast=False to disable this error"
                )
                raise exceptions.CastValueFailure(msg)

        meta = SimpleNamespace(
            cast=cast,
            default=default,
            loaders=loaders,
            value=result,
            cast_from=cast_from,
            loader_from=loader_from,
            results_from=results_from,
            default_from=default_from,
        )

        return result, meta


class FieldConf(Field):
    """A field that represents a nested configuration.

    This field type allows for hierarchical configuration structures by containing
    another configuration class as its value.

    Attributes:
        children_class: The configuration class to use for nested values.
    """

    def __init__(
        self,
        children_class,
        key: str = None,
        **kwargs,
    ):
        """Initialize a nested configuration field.

        Args:
            children_class: The configuration class to use for nested values.
            key: Name of the value used in file or environment variable.
            **kwargs: Additional arguments passed to the parent Field class.
        """
        super().__init__(key, **kwargs)
        self.children_class = children_class


class FieldBool(Field):
    """A field that stores and validates boolean values.

    Uses the AsBoolean cast to convert various string representations
    to boolean values (e.g., 'yes'/'no', 'true'/'false', '1'/'0').

    Attributes:
        cast: Set to AsBoolean() for automatic type conversion.
    """

    cast = as_boolean


class FieldString(Field):
    """A field that stores string values.

    Ensures values are stored as strings, converting other types
    if necessary using Python's built-in str() function.

    Attributes:
        cast: Set to str for automatic type conversion.
    """

    cast = str


class FieldInt(Field):
    """A field that stores integer values.

    Uses the AsInt cast to convert string representations to integers,
    raising an error if the conversion fails.

    Attributes:
        cast: Set to AsInt() for automatic type conversion.
    """

    cast = as_int


class FieldFloat(Field):
    """A field that stores floating-point values.

    Uses Python's built-in float() function to convert values,
    raising an error if the conversion fails.

    Attributes:
        cast: Set to float for automatic type conversion.
    """

    cast = float


class FieldOption(Field):
    """A field that validates values against a predefined set of options.

    This field ensures that values are one of a predefined set of options,
    optionally providing a default if an invalid option is given.

    Attributes:
        cast: Set to AsOption for option validation and conversion.
    """

    cast = as_option

    def __init__(
        self,
        options,
        default_option=FAIL,
        key: str = None,
        **kwargs,
    ):
        """Initialize an option field.

        Args:
            options: Dictionary mapping valid input values to their corresponding options.
            default_option: The option to use when an invalid value is provided.
                If set to FAIL, raises an error for invalid values.
            key: Name of the value used in file or environment variable.
            **kwargs: Additional arguments passed to the parent Field class.

        Raises:
            AssertionError: If options is not a dictionary.
        """
        assert isinstance(options, dict), f"Expected a dict, got: {options}"
        self.cast = AsOption(options, default_option=default_option)
        super().__init__(key, **kwargs)


class FieldDict(Field):
    """A field that stores dictionary values.

    Uses the AsDict cast to ensure values are proper dictionaries,
    with support for converting mapping objects.

    Attributes:
        cast: Set to AsDict() for automatic type conversion.
    """

    cast = as_dict


class FieldList(Field):
    """A field that stores list values.

    Uses the AsList cast to convert various inputs to lists, including:
    - Comma-separated strings
    - Other sequence types
    - Empty values to empty lists

    Attributes:
        cast: Set to AsList() for automatic type conversion.
    """

    cast = as_list


class FieldTuple(Field):
    """A field that stores tuple values.

    Uses the AsTuple cast to convert various inputs to tuples, including:
    - Comma-separated strings
    - Other sequence types
    - Empty values to empty tuples

    Attributes:
        cast: Set to AsTuple() for automatic type conversion.
    """

    cast = as_tuple


# Compatibility with classyconf
Value = Field  # Alias for backward compatibility



"Support value casting"

# pylint: disable=too-few-public-methods, too-many-return-statements

import ast
from collections.abc import Mapping, Sequence

from .common import FAIL  # , NOT_SET, UNSET_ARG
from .exceptions import InvalidCastConfiguration


class AbstractCast:
    """Base class for all cast operations.

    This abstract class defines the interface that all cast implementations must follow.
    Subclasses must implement the __call__ method to perform the actual casting operation.
    """

    def __call__(self, value):
        raise NotImplementedError()  # pragma: no cover


class AsBoolean(AbstractCast):
    """Cast a value to a boolean using predefined string mappings.

    Converts various string representations to boolean values using a configurable
    mapping dictionary. By default, supports common boolean string representations
    like 'true', 'yes', 'on', '1' for True and their counterparts for False.

    Args:
        values (dict, optional): A dictionary mapping strings to boolean values.
            If provided, updates the default mapping dictionary.

    Raises:
        InvalidCastConfiguration: If the input value cannot be cast to a boolean.
    """

    default_values = {
        "1": True,
        "true": True,
        "yes": True,
        "y": True,
        "on": True,
        "t": True,
        "0": False,
        "false": False,
        "no": False,
        "n": False,
        "off": False,
        "f": False,
    }

    def __init__(self, values=None):
        self.values = self.default_values.copy()
        if isinstance(values, dict):
            self.values.update(values)

    def __call__(self, value):
        try:
            return self.values[str(value).lower()]
        except KeyError as err:
            raise InvalidCastConfiguration(
                f"Error casting value {value} to boolean"
            ) from err


class AsString(AbstractCast):
    """Cast a value to a string.

    Attempts to convert the input value to a string using Python's built-in str() function.

    Raises:
        InvalidCastConfiguration: If the value cannot be converted to a string.
    """

    def __call__(self, value):
        if isinstance(value, str):
            return str(value)
        if not value:
            return ""
        return str(value)


class AsInt(AbstractCast):
    """Cast a value to an integer.

    Attempts to convert the input value to an integer using Python's built-in int() function.

    Raises:
        InvalidCastConfiguration: If the value cannot be converted to an integer.
    """

    def __call__(self, value):
        try:
            return int(value)
        except (ValueError, TypeError) as err:
            # TOFIX: Raise or report unset ?
            # return NOT_SET
            raise InvalidCastConfiguration(
                f"Error casting value {value} to int"
            ) from err


class AsList(AbstractCast):
    """Cast a value to a list with support for delimited strings.

    Converts various input types to a list:
    - Empty values become empty lists
    - Strings are split by delimiter, with support for quoted elements
    - Sequences are converted directly to lists

    Args:
        delimiter (str, optional): The character used to split strings. Defaults to ','.
        quotes (str, optional): String containing valid quote characters. Defaults to '"\''.

    Examples:
        >>> cast = AsList()
        >>> cast('a,b,c')  # Returns: ['a', 'b', 'c']
        >>> cast('"a,b",c')  # Returns: ['a,b', 'c']
        >>> cast(['a', 'b'])  # Returns: ['a', 'b']
    """

    def __init__(self, delimiter=",", quotes="\"'"):
        self.delimiter = delimiter
        self.quotes = quotes

    def cast(self, sequence):
        "Cast to correct type"
        return list(sequence)

    def __call__(self, value):
        return self._parse(value)

    def _parse(self, value):

        if not value:
            # print ("PARSE AS EMPTY", value)
            return self.cast([])

        if isinstance(value, str):
            # print ("PARSE AS STRING", value)
            return self._parse_string(value)

        if isinstance(value, Sequence):
            # print ("PARSE AS LIST", value)
            return self.cast(value)
        if isinstance(value, Mapping):
            assert False, f"TOFIX: Unsupported type dict, {value}"

        assert False

    def _parse_string(self, string):
        elements = []
        element = []
        quote = ""
        for char in string:
            # open quote
            if char in self.quotes and not quote:
                quote = char
                element.append(char)
                continue

            # close quote
            if char in self.quotes and char == quote:
                quote = ""
                element.append(char)
                continue

            if quote:
                element.append(char)
                continue

            if char == self.delimiter:
                elements.append("".join(element))
                element = []
                continue

            element.append(char)

        # remaining element
        if element:
            elements.append("".join(element))

        return self.cast(e.strip() for e in elements)


class AsTuple(AsList):
    """Cast a value to a tuple.

    Inherits from AsList but converts the final result to a tuple instead of a list.
    Accepts the same arguments and follows the same parsing rules as AsList.

    Args:
        delimiter (str, optional): The character used to split strings. Defaults to ','.
        quotes (str, optional): String containing valid quote characters. Defaults to '"\''.
    """

    def cast(self, sequence):
        return tuple(sequence)


class AsDict(AbstractCast):
    """Cast a value to a dictionary.

    Currently supports:
    - Empty values become empty dictionaries
    - Mapping objects are converted directly to dictionaries

    Args:
        delimiter (str, optional): Reserved for future string parsing. Defaults to ','.
        quotes (str, optional): Reserved for future string parsing. Defaults to '"\''.

    Note:
        String parsing is not yet implemented.
    """

    def __init__(self, delimiter=",", quotes="\"'"):
        self.delimiter = delimiter
        self.quotes = quotes

    def cast(self, sequence):
        "Cast value"
        return dict(sequence)

    def __call__(self, value):
        return self._parse(value)

    def _parse(self, value):
        "Internal helper to parse values"

        if not value:
            # print ("PARSE AS EMPTY", value)
            return self.cast({})

        if isinstance(value, str):
            assert False, "String  parsing is not implemeted yet"
            # print ("PARSE AS STRING", value)
            # return self._parse_string(value)

        if isinstance(value, Mapping):
            # print ("PARSE AS LIST", value)
            return self.cast(value)
        if isinstance(value, Sequence):
            assert False, f"TOFIX: Unsupported type list, {value}"

        assert False


class AsOption(AbstractCast):
    """Cast a value by selecting from predefined options.

    Maps input values to predefined options using a dictionary mapping.
    Optionally supports a default option when the input doesn't match any defined option.

    Args:
        options (dict): A dictionary mapping input values to their corresponding options.
        default_option (any, optional): The key to use when the input value isn't found.
            If FAIL (default), raises an exception for invalid inputs.

    Raises:
        InvalidCastConfiguration: If the input value is not in options and no valid
            default_option is provided.

    Example:
        >>> cast = AsOption({'dev': ['debug'], 'prod': ['optimize']}, 'dev')
        >>> cast('prod')  # Returns: ['optimize']
        >>> cast('invalid')  # Returns: ['debug'] (default option)
    """

    def __init__(self, options, default_option=FAIL):
        self.options = options
        self.default_option = default_option

    def __call__(self, value):
        try:
            return self.options[value]
        except KeyError as err:

            # Raise error if no default
            default_option = self.default_option
            if default_option is FAIL:
                raise InvalidCastConfiguration(f"Invalid option {value}") from err

            # Look for default
            if not default_option in self.options:
                raise InvalidCastConfiguration(
                    f"Invalid default option {value}: does not exists: {default_option}"
                ) from err

            # if isinstance(default, str):
            return self.options[default_option]
            # if ret is NOT_SET:
            #     raise InvalidCastConfiguration("Invalid default option {!r}".format(value))

            # return ret


class AsIdentity(AbstractCast):
    """Return the input value unchanged.

    A no-operation cast that simply returns the input value without modification.
    Useful as a default cast or when you need to maintain the original type.
    """

    def __call__(self, value):
        return value


class AsBest(AbstractCast):
    """Cast a value to its most appropriate Python type.

    Attempts to intelligently determine and convert the input value to the most
    suitable Python type by trying different conversions in order:
    1. None for null/empty values
    2. Boolean for true/false-like strings
    3. Integer for numeric strings without decimals
    4. Float for numeric strings with decimals
    5. List for comma-separated strings or sequence-like inputs
    6. Dict for mapping-like inputs
    7. Original string if no other conversion succeeds

    Examples:
        >>> cast = AsBest()
        >>> cast('123')  # Returns: 123 (int)
        >>> cast('3.14')  # Returns: 3.14 (float)
        >>> cast('true')  # Returns: True (bool)
        >>> cast('a,b,c')  # Returns: ['a', 'b', 'c'] (list)
    """

    def __init__(self, delimiter=",", quotes="\"'"):
        self.delimiter = delimiter
        self.list_caster = AsList(delimiter=delimiter, quotes=quotes)
        self.dict_caster = AsDict(delimiter=delimiter, quotes=quotes)

    def __call__(self, value):
        if not value:
            return None

        # Try boolean first for true/false strings
        if isinstance(value, str):

            # Try integer
            try:
                if "." not in value:
                    return int(value)
            except ValueError:
                pass

            # Try float
            try:
                return float(value)
            except ValueError:
                pass

            # Try list (if contains delimiter)
            if self.delimiter and self.delimiter in value:
                try:
                    return self.list_caster(value)
                except (InvalidCastConfiguration, AssertionError):
                    pass

        # Try dict for mapping types
        if isinstance(value, Mapping):
            try:
                return self.dict_caster(value)
            except (InvalidCastConfiguration, AssertionError):
                pass

        # Try list for sequence types
        if isinstance(value, Sequence) and not isinstance(value, str):
            try:
                return self.list_caster(value)
            except (InvalidCastConfiguration, AssertionError):
                pass

        # Return as is if no conversion succeeded
        return value


evaluate = ast.literal_eval



"Loaders library"

# pylint: disable=unnecessary-dunder-call, unused-argument


import os
from configparser import ConfigParser, MissingSectionHeaderError, NoOptionError
from glob import glob

from .common import NOT_SET
from .exceptions import InvalidConfigurationFile, InvalidPath, MissingSettingsSection
from .parsers import EnvFileParser


class EnvPrefix:
    """
    A utility class for namespacing environment variables with a prefix.

    Since the environment is a global dictionary, it is a good practice to
    namespace your settings by using a unique prefix like ``MY_APP_``.

    Args:
        prefix (str): The prefix to prepend to environment variable names. Defaults to "".

    Example:
        >>> prefix = EnvPrefix("MYAPP_")
        >>> prefix("database_url")
        'MYAPP_DATABASE_URL'
    """

    def __init__(self, prefix=""):
        self.prefix = prefix

    def __call__(self, value):
        """
        Transform a configuration key into an environment variable name.

        Args:
            value (str): The configuration key to transform.

        Returns:
            str: The environment variable name with the prefix applied and converted to uppercase.
        """
        value = str(value)
        return f"{self.prefix}{value.upper()}"

    def __repr__(self):
        return f'{self.__class__.__name__}("{self.prefix}")'


def _get_args(parser):
    """
    Convert arguments extracted from an ArgumentParser to a dictionary.

    Args:
        parser (argparse.ArgumentParser): An ArgumentParser instance.

    Returns:
        dict: Dictionary containing the parsed CLI arguments, excluding those with NOT_SET values.
    """
    args = vars(parser.parse_args()).items()
    return {key: val for key, val in args if not isinstance(val, type(NOT_SET))}


class AbstractConfigurationLoader:
    """
    Abstract base class for configuration loaders.

    This class defines the interface that all configuration loaders must implement.
    Configuration loaders are responsible for loading configuration values from
    different sources (files, environment, etc.).
    """

    def __repr__(self):
        raise NotImplementedError()  # pragma: no cover

    def __contains__(self, item):
        """Check if a configuration key exists in this loader."""
        raise NotImplementedError()  # pragma: no cover

    def __getitem__(self, item):
        """Retrieve a configuration value by key."""
        raise NotImplementedError()  # pragma: no cover

    def check(self):
        """
        Verify if the configuration source is valid and accessible.

        Returns:
            bool: True if the configuration source is valid, False otherwise.
        """
        return True

    def reset(self):
        """Reset the loader's internal state."""

    # pylint: disable=unused-argument
    def contains(self, config, item):
        """
        Check if a configuration key exists in this loader.

        Args:
            config: The configuration object.
            item (str): The configuration key to check.

        Returns:
            bool: True if the key exists, False otherwise.
        """
        return self.__contains__(item)
        # raise NotImplementedError()  # pragma: no cover

    def getitem(self, config, item):
        """
        Retrieve a configuration value by key.

        Args:
            config: The configuration object.
            item (str): The configuration key to retrieve.

        Returns:
            The configuration value for the given key.

        Raises:
            KeyError: If the key doesn't exist.
        """
        return self.__getitem__(item)
        # raise NotImplementedError()  # pragma: no cover


class CommandLine(AbstractConfigurationLoader):
    """
    Configuration loader that extracts settings from command line arguments.

    This loader uses an argparse.ArgumentParser to extract configuration values
    from command line arguments.

    Args:
        parser (argparse.ArgumentParser): The parser instance to extract variables from.
        get_args (callable): Optional function to extract args from the parser.
                           Defaults to the get_args function.
    """

    # noinspection PyShadowingNames
    def __init__(self, parser, get_args=_get_args):
        """
        :param parser: An `argparse` parser instance to extract variables from.
        :param function get_args: A function to extract args from the parser.
        :type parser: argparse.ArgumentParser
        """
        self.parser = parser
        self.configs = get_args(self.parser)

    def __repr__(self):
        return f"{self.__class__.__name__}(parser={self.parser})"

    def __contains__(self, item):
        """Check if a configuration key exists in the parsed arguments."""
        return item in self.configs

    def __getitem__(self, item):
        """Retrieve a configuration value from the parsed arguments."""
        return self.configs[item]


class IniFile(AbstractConfigurationLoader):
    """
    Configuration loader that reads settings from an INI/CFG file.

    This loader reads configuration from a specified section in an INI-style
    configuration file.

    Args:
        filename (str): Path to the .ini/.cfg file.
        section (str): Section name inside the config file. Defaults to "settings".
        keyfmt (callable): Optional function to pre-format variable names.

    Raises:
        InvalidConfigurationFile: If the file is not a valid INI file.
        MissingSettingsSection: If the specified section is not found in the file.
    """

    def __init__(self, filename, section="settings", keyfmt=lambda x: x):
        self.filename = filename
        self.section = section
        self.keyfmt = keyfmt
        self.parser = ConfigParser(allow_no_value=True)
        self._initialized = False

    def __repr__(self):
        return f'{self.__class__.__name__}("{self.filename}")'

    def _parse(self):
        """
        Parse the INI file if not already parsed.

        Raises:
            InvalidConfigurationFile: If the file is not a valid INI file.
            MissingSettingsSection: If the specified section is not found.
        """
        if self._initialized:
            return

        with open(self.filename, encoding="utf-8") as inifile:
            try:
                self.parser.read_file(inifile)
            except (UnicodeDecodeError, MissingSectionHeaderError) as err:
                raise InvalidConfigurationFile() from err

        if not self.parser.has_section(self.section):
            raise MissingSettingsSection(
                f"Missing [{self.section}] section in {self.filename}"
            )

        self._initialized = True

    def check(self):
        """
        Verify if the INI file exists and is valid.

        Returns:
            bool: True if the file exists and is valid, False otherwise.
        """
        try:
            self._parse()
        except (FileNotFoundError, InvalidConfigurationFile, MissingSettingsSection):
            return False

        return super().check()

    def __contains__(self, item):
        """Check if a configuration key exists in the INI file section."""
        if not self.check():
            return False

        return self.parser.has_option(self.section, self.keyfmt(item))

    def __getitem__(self, item):
        """
        Retrieve a configuration value from the INI file section.

        Args:
            item (str): The configuration key to retrieve.

        Returns:
            str: The configuration value.

        Raises:
            KeyError: If the key doesn't exist in the section.
        """
        if not self.check():
            raise KeyError(f"{item}")

        try:
            return self.parser.get(self.section, self.keyfmt(item))
        except NoOptionError as err:
            raise KeyError(f"{item}") from err

    def reset(self):
        """Reset the parser's initialization state."""
        self._initialized = False


class Environment(AbstractConfigurationLoader):
    """
    Configuration loader that reads settings from environment variables.

    This loader retrieves configuration values from the system's environment
    variables (os.environ), optionally transforming the keys using a formatting
    function.

    Args:
        keyfmt (callable): Optional function to pre-format variable names.
                          Defaults to EnvPrefix() which converts keys to uppercase.

    Example:
        >>> env = Environment(keyfmt=EnvPrefix("MYAPP_"))
        >>> os.environ["MYAPP_DEBUG"] = "true"
        >>> "debug" in env  # True
        >>> env["debug"]  # "true"
    """

    def __init__(self, keyfmt=EnvPrefix(), prefix: str = None, sep: str = "__"):
        """
        :param function keyfmt: A function to pre-format variable names.
        """
        self.keyfmt = keyfmt
        self.prefix = prefix
        self.sep = sep

    def __repr__(self):
        return f"{self.__class__.__name__}(keyfmt={self.keyfmt})"

    def __contains__(self, item: str) -> bool:
        """Check if an environment variable exists for the given key."""
        return self.keyfmt(item) in os.environ

    def __getitem__(self, item: str):
        """
        Retrieve a configuration value from environment variables.

        Args:
            item (str): The configuration key to retrieve.

        Returns:
            str: The environment variable value.

        Raises:
            KeyError: If the environment variable doesn't exist.
        """
        return os.environ[self.keyfmt(item)]

    def get_env_name(self, config, item: str) -> str:
        "Return the envirnonment name for a given key in config"

        # Check if environment is enabled, and parents
        pkey_enabled = config.query_inst_cfg("env_enabled", default=True)
        parents = []
        if pkey_enabled:
            parents = config.get_hierarchy()  # [:-1]

        # For each parents, from bottom to top
        parents_keys = []
        for parent in parents:

            # Fetch pkey info
            pkey = parent.query_inst_cfg("env_name", default=parent.key)
            pkey_prefix = parent.query_inst_cfg("env_prefix", default="")
            pkey_pattern = parent.query_inst_cfg("env_pattern", default="{prefix}{key}")

            # Check if root:
            pkey_prefix = pkey_prefix or ""
            # pylint: disable=protected-access
            if parent._parent is None and pkey is None:
                # Is probably root, then we have to guess the name if not explicit
                pkey = ""
                if not pkey_prefix:
                    pkey = parent.__class__.__name__
                    pkey_prefix = ""

            # Render format
            pfinal = pkey_pattern.format(
                prefix=pkey_prefix,
                key=pkey,
            )
            pfinal = [str(x) for x in [pkey_prefix, pkey] if x]
            pfinal = "_".join(pfinal)
            parents_keys.append(pfinal)

            # Quit if object does not want parents
            pkey_parents = parent.query_inst_cfg("env_parents", default=True)
            if pkey_parents is False:
                break

        # Reverse key order, from top to bottom, transform to string, and then uppsercase
        parents_keys = list(reversed(parents_keys))
        parents_keys.append(item)
        parents_keys = [str(x) for x in parents_keys if x]
        # print(parents_keys)
        fkey = self.sep.join(parents_keys).upper()

        return fkey

    def getitem(self, config, item: str):
        """
        Retrieve a configuration value by key.

        Args:
            config: The configuration object.
            item (str): The configuration key to retrieve.

        Returns:
            The configuration value for the given key.

        Raises:
            KeyError: If the key doesn't exist.
        """

        fkey = self.get_env_name(config, item)
        # print("==== LOOK  ENV", fkey)
        ret = os.environ[fkey]
        # print("==== FOUND ENV", fkey, "=", ret)

        ret = os.path.expandvars(ret)
        ret = os.path.expanduser(ret)
        return ret

        # return self.__getitem__(item)
        # raise NotImplementedError()  # pragma: no cover


class EnvFile(AbstractConfigurationLoader):
    """
    Configuration loader that reads settings from a .env file.

    This loader reads configuration values from a file in environment variable
    format (KEY=value).

    Args:
        filename (str): Path to the .env file. Defaults to ".env".
        keyfmt (callable): Optional function to pre-format variable names.
                          Defaults to EnvPrefix() which converts keys to uppercase.

    Example:
        >>> env = EnvFile(".env", keyfmt=EnvPrefix("MYAPP_"))

        With .env containing: DEBUG=true

        >>> "debug" in env

        Returns `True`

        >>> env["debug"]

        Return `"true"`
    """

    def __init__(self, filename=".env", keyfmt=EnvPrefix()):
        self.filename = filename
        self.keyfmt = keyfmt
        self.configs = None

    def __repr__(self):
        return f'{self.__class__.__name__}("{self.filename}")'

    def _parse(self):
        """
        Parse the .env file if not already parsed.

        Raises:
            FileNotFoundError: If the file doesn't exist or can't be read.
        """
        if self.configs is not None:
            return

        self.configs = {}
        with open(self.filename, encoding="utf-8") as envfile:
            self.configs.update(EnvFileParser(envfile).parse_config())

    def check(self):
        """
        Verify if the .env file exists and is valid.

        Returns:
            bool: True if the file exists and is valid, False otherwise.
        """
        if not os.path.isfile(self.filename):
            return False

        try:
            self._parse()
        except FileNotFoundError:
            return False

        return super().check()

    def __contains__(self, item):
        """Check if a configuration key exists in the .env file."""
        if not self.check():
            return False

        return self.keyfmt(item) in self.configs

    def __getitem__(self, item):
        """
        Retrieve a configuration value from the .env file.

        Args:
            item (str): The configuration key to retrieve.

        Returns:
            str: The configuration value.

        Raises:
            KeyError: If the key doesn't exist in the file.
        """
        if not self.check():
            raise KeyError(f"{item}")

        return self.configs[self.keyfmt(item)]

    def reset(self):
        """Reset the parsed configuration cache."""
        self.configs = None


class RecursiveSearch(AbstractConfigurationLoader):
    """
    Configuration loader that recursively searches for configuration files.

    This loader looks for configuration files in the current directory and all parent
    directories up to a specified root path. It supports multiple file types and loaders.

    Args:
        starting_path (str, optional): The path to begin looking for configuration files.
        filetypes (tuple): Tuple of (pattern, loader_class) pairs defining which files
                          to look for and how to load them. Defaults to .env, .ini, and .cfg files.
        root_path (str): The path where the search will stop. Defaults to "/".

    Example:
        >>> search = RecursiveSearch("/home/user/project")

        Will look for .env, .ini, and .cfg files in:
        /home/user/project
        /home/user
        /home
        /
    """

    def __init__(
        self,
        starting_path=None,
        filetypes=((".env", EnvFile), (("*.ini", "*.cfg"), IniFile)),
        root_path="/",
    ):
        self.root_path = os.path.realpath(root_path)
        self._starting_path = self.root_path

        if starting_path:
            self.starting_path = starting_path

        self.filetypes = filetypes
        self._config_files = None

    @property
    def starting_path(self):
        """Get the current starting path for the search."""
        return self._starting_path

    @starting_path.setter
    def starting_path(self, path):
        """
        Set the starting path for the search.

        Args:
            path (str): The new starting path.

        Raises:
            InvalidPath: If the path is invalid or outside the root path.
        """
        if not path:
            raise InvalidPath("Invalid starting path")

        path = os.path.realpath(os.path.abspath(path))
        if not path.startswith(self.root_path):
            raise InvalidPath("Invalid root path given")
        self._starting_path = path

    @staticmethod
    def get_filenames(path, patterns):
        """
        Get all filenames in a directory matching the given patterns.

        Args:
            path (str): Directory to search in.
            patterns (str or tuple): Glob pattern(s) to match against.

        Returns:
            list: List of matching filenames.
        """
        filenames = []
        if isinstance(patterns, str):
            patterns = (patterns,)

        for pattern in patterns:
            filenames += glob(os.path.join(path, pattern))
        return filenames

    def _scan_path(self, path):
        """
        Scan a directory for configuration files.

        Args:
            path (str): Directory to scan.

        Returns:
            list: List of configuration loader instances for found files.
        """
        config_files = []

        for patterns, Loader in self.filetypes:
            for filename in self.get_filenames(path, patterns):
                try:
                    loader = Loader(filename=filename)
                    if not loader.check():
                        continue
                    config_files.append(loader)
                except InvalidConfigurationFile:
                    continue

        return config_files

    def _discover(self):
        """Discover all configuration files in the search path hierarchy."""
        self._config_files = []

        path = self.starting_path
        while True:
            if os.path.isdir(path):
                self._config_files += self._scan_path(path)

            if path == self.root_path:
                break

            path = os.path.dirname(path)

    @property
    def config_files(self):
        """
        Get all discovered configuration files.

        Returns:
            list: List of configuration loader instances.
        """
        if self._config_files is None:
            self._discover()

        return self._config_files

    def __contains__(self, item):
        """Check if a configuration key exists in any of the discovered files."""
        for config_file in self.config_files:
            if item in config_file:
                return True
        return False

    def __getitem__(self, item):
        """
        Retrieve a configuration value from the first file that contains it.

        Args:
            item (str): The configuration key to retrieve.

        Returns:
            str: The configuration value.

        Raises:
            KeyError: If the key doesn't exist in any configuration file.
        """
        for config_file in self.config_files:
            try:
                return config_file[item]
            except KeyError:
                continue

        raise KeyError(f"{item}")

    def __repr__(self):
        return f"{self.__class__.__name__}(root_path={self.root_path})"

    def reset(self):
        """Reset the discovered configuration files cache."""
        self._config_files = None


class Dict(AbstractConfigurationLoader):
    """
    Configuration loader that uses a dictionary as the configuration source.

    This loader is useful for providing hardcoded default values or for testing.

    Args:
        values_mapping (dict): A dictionary of configuration key-value pairs.

    Example:
        >>> config = Dict({"debug": "true", "port": "8000"})
        >>> "debug" in config  # True
        >>> config["port"]  # "8000"
    """

    def __init__(self, values_mapping):
        self.values_mapping = values_mapping

    def __repr__(self):
        return f"{self.__class__.__name__}({self.values_mapping})"

    def __contains__(self, item):
        """Check if a configuration key exists in the dictionary."""
        return item in self.values_mapping

    def __getitem__(self, item):
        """
        Retrieve a configuration value from the dictionary.

        Args:
            item (str): The configuration key to retrieve.

        Returns:
            The configuration value.

        Raises:
            KeyError: If the key doesn't exist in the dictionary.
        """
        return self.values_mapping[item]


class _Value(AbstractConfigurationLoader):
    """
    Internal configuration loader for handling single values.

    This is an internal class used for managing individual configuration values
    within the configuration system.

    Args:
        values_mapping (dict): A dictionary containing the configuration value.
    """

    def __init__(self, values_mapping):
        self.values_mapping = values_mapping

    def __repr__(self):
        return f"{self.__class__.__name__}({self.values_mapping})"

    def __contains__(self, item):
        """Check if the value exists for the given key."""
        return item in self.values_mapping

    def __getitem__(self, item):
        """Retrieve the value for the given key."""
        return self.values_mapping[item]

    def getitem(self, config, item, value=NOT_SET, **kwargs):
        """
        Get or set a configuration value.

        Args:
            config: The configuration object.
            item (str): The configuration key.
            value: The value to set (optional).
            **kwargs: Additional arguments.

        Returns:
            The configuration value.
        """
        if value is not NOT_SET:
            self.values_mapping = value
        return self.__getitem__(item)


class _Value22(AbstractConfigurationLoader):
    """
    Alternative internal configuration loader for handling single values.

    This is an internal class used for managing individual configuration values
    with slightly different behavior than _Value.

    Args:
        values_mapping (dict): A dictionary containing the configuration value.
    """

    def __init__(self, values_mapping):
        self.values_mapping = values_mapping

    def __repr__(self):
        return f"{self.__class__.__name__}({self.values_mapping})"

    def __contains__(self, item):
        """Check if the value exists for the given key."""
        return item in self.values_mapping

    def __getitem__(self, item):
        """Retrieve the value for the given key."""
        return self.values_mapping[item]

    def getitem(self, config, item, value=NOT_SET, **kwargs):
        """
        Get or set a configuration value.

        Args:
            config: The configuration object.
            item (str): The configuration key.
            value: The value to set (optional).
            **kwargs: Additional arguments.

        Returns:
            The value parameter if provided, otherwise the stored value.
        """
        return value



"Main configuratio  class"


# pylint: disable=unused-argument, too-few-public-methods, too-many-instance-attributes, use-dict-literal, protected-access, too-many-branches

import copy
import logging
from collections import OrderedDict

# from collections import Mapping, Sequence
from collections.abc import Mapping, Sequence

# pylint: disable=unused-import
from pprint import pprint

from superconf import exceptions

from .common import NOT_SET, UNSET_ARG
from .fields import Field, FieldConf
from .loaders import Environment

# from types import SimpleNamespace
# from typing import Callable


logger = logging.getLogger(__name__)


# ====================================
# Configuration Child
# ====================================


class Node:
    """Base class for configuration objects providing core configuration query functionality.

    This class implements the basic configuration query mechanisms used by all configuration
    classes. It supports querying configuration values from various sources including
    instance attributes, class Meta attributes, and parent configurations.
    """

    class Meta:
        """Class to store class-level configuration overrides."""

    def __init__(self, key=None, value=NOT_SET, parent=None):
        """Initialize a configuration base instance.

        Args:
            key: The configuration key name
            value: The configuration value (defaults to NOT_SET)
            parent: Parent configuration object if this is a child config
        """
        self.key = key
        self._parent = parent
        self._value = value
        self._cache = True  # TOFIX

    # Instance config management
    # ----------------------------

    def query_inst_cfg(self, *args, cast=None, report=False, **kwargs):
        """Query instance configuration with optional type casting.

        Args:
            *args: Variable length argument list passed to _query_inst_cfg
            cast: Optional type to cast the result to
            **kwargs: Arbitrary keyword arguments passed to _query_inst_cfg

        Returns:
            The configuration value, optionally cast to the specified type
        """
        out, _report = self._query_inst_cfg(*args, **kwargs)
        logger.debug("Node config query for %s.%s=%s", self, args[0], out)

        if isinstance(out, (dict, list)):
            out = copy.copy(out)

        if cast is not None:
            # Try to cast if asked
            if not out:
                out = cast()
            assert isinstance(
                out, cast
            ), f"Wrong type for config {self}, expected {cast}, got: {type(out)} {out}"
        if report:
            return out, _report
        return out

    # @classmethod
    # def _query_cls_cfg(cls, *args, **kwargs):
    #     "Temporary class method"
    #     out = cls._query_inst_cfg(cls, *args, **kwargs)
    #     if isinstance(out, (dict, list)):
    #         out = copy.copy(out)
    #     return out

    def _query_inst_cfg(self, name, override=None, default=UNSET_ARG):
        """Internal method to query instance configuration from various sources.

        Searches for configuration values in the following order:
        1. Dictionary override if provided
        2. Instance attribute with _name prefix
        3. Class Meta attribute
        4. Instance attribute with meta__ prefix
        5. Default value if provided

        Args:
            name: Configuration setting name to query
            override: Optional dictionary of override values
            parents: Whether to check parent configurations
            default: Default value if setting is not found

        Returns:
            Tuple of (value, query_sources) where query_sources is a list of searched locations

        Raises:
            UnknownSetting: If the setting is not found and no default is provided
        """
        query_from = []

        # Fetch from dict override, if provided
        if isinstance(override, dict):
            val = override.get(name, NOT_SET)
            if val is not NOT_SET:
                query_from.append(f"dict_override:{name}")
                return val, query_from

        # Fetch from self._NAME
        # Good for initial setup, if write mode is required
        val = getattr(self, f"_{name}", NOT_SET)
        if val is not NOT_SET:
            query_from.append(f"self_attr:_{name}")
            return val, query_from

        # Python class params
        # Good for class overrides
        cls = self
        if hasattr(cls, "Meta"):
            val = getattr(cls.Meta, name, NOT_SET)
            if val is not NOT_SET:
                query_from.append(f"self_meta:Meta.{name}")
                # print ("SELF CLASS Meta retrieval for: {cls}" , name, val)
                return val, query_from

        # Fetch from self.meta__NAME
        # Python class inherited params (good for defaults)
        val = getattr(self, f"meta__{name}", NOT_SET)
        if val is not NOT_SET:
            query_from.append(f"self_attr:meta__{name}")
            return val, query_from

        if default is not UNSET_ARG:
            query_from.append("default_arg")
            return default, query_from

        msg = (
            f"Setting '{name}' has not been declared before being used"
            f" in '{repr(self)}', tried to query: {query_from}"
        )
        raise exceptions.UnknownSetting(msg)

    def query_cfg(self, name, include_self=True, report=False, **kwargs):
        "Temporary wrapper"
        return self.query_parent_cfg(name, include_self=True, report=report, **kwargs)

    # pylint: disable=too-many-arguments, too-many-positional-arguments
    def query_parent_cfg(
        self,
        name,
        as_subkey=False,
        cast=None,
        default=UNSET_ARG,
        include_self=False,
        report=False,
    ):
        """Query configuration from parent object.

        Args:
            name: Configuration setting name to query
            as_subkey: If True and parent value is dict, get self.key from it
            cast: Optional type to cast the result to
            default: Default value if setting is not found

        Returns:
            The configuration value from the parent, optionally cast to specified type

        Raises:
            UnknownSetting: If no parent exists and no default is provided
        """

        # Fast exit or raise exception
        if not self._parent and include_self is False:
            if default is not UNSET_ARG:
                if report:
                    return default, "No parents, return default value"
                return default
            msg = (
                f"Setting '{name}' has not been declared in hierarchy of '{repr(self)}'"
            )
            raise exceptions.UnknownSetting(msg)

        # Check parents
        _report = []
        parents = self.get_hierarchy()
        if include_self is False:
            parents = parents[1:]
        out = NOT_SET
        for parent in parents:
            _report.append(f"Check '{name}' in parent {parent}")

            if report:
                out, _report2 = parent.query_inst_cfg(
                    name, default=NOT_SET, report=True
                )
                _report.append(_report2)
            else:
                out = parent.query_inst_cfg(name, default=NOT_SET)

            # If a value is found, then scan it
            if out is not NOT_SET:
                _report.append(f"Found '{name}' in parent {parent}= {out}")

                # Ckeck subkey
                if as_subkey is True:
                    if isinstance(out, dict):
                        out = out.get(self.key, NOT_SET)
                    elif isinstance(out, list):
                        assert isinstance(self.key, int), f"Got: {self.key}"
                        out = out[self.key]
                    else:
                        out = NOT_SET
                _report.append(f"Found2 '{name}' in parent {parent}= {out}")

            # Don't ask more parents if value is found
            if out is not NOT_SET:
                break

        if out is NOT_SET:
            _report.append(f"NotFound '{name}' in parent: {parent}")

        if cast is not None:
            # Try to cast if asked
            if not out:
                out = cast()
            assert isinstance(
                out, cast
            ), f"Wrong type for config {name}, expected {cast}, got: {type(out)} {out}"

        if out is NOT_SET and default is UNSET_ARG:
            msg = (
                f"Setting '{name}' has not been declared in hierarchy of '{repr(self)}'"
            )
            raise exceptions.UnknownSetting(msg)

        if report:
            return out, _report
        return out

    def get_hierarchy(self):
        "Return a list of parents NEW VERSION"
        out = [self]

        target = self
        while target._parent is not None and target._parent not in out:
            target = target._parent
            out.append(target)

        return out


# ====================================
# Configuration Container
# ====================================


class Store(Node):
    """Base configuration container class that manages configuration fields and values.

    This class extends Node to provide field management, value caching,
    and dynamic child configuration creation capabilities.
    """

    _declared_values = {}

    def __init__(self, *, key=None, value=NOT_SET, parent=None, meta=None, **kwargs):
        """Initialize a configuration container.

        Args:
            key: Configuration key name
            value: Initial configuration value
            parent: Parent configuration object
            meta: Optional meta configuration
            **kwargs: Additional configuration options
        """

        # super(Store, self).__init__(key=key, value=value, parent=parent)
        super().__init__(key=key, value=value, parent=parent)

        # As this can be updated during runtime ...
        # self._declared_values = self._declared_values
        # self._declared_values = dict()
        self._cached_values = {}

        kwargs.update(
            dict(
                key=key,
                # loaders=loaders,
                # cache=cache,
                parent=parent,
            )
        )

        # self._loaders = NOT_SET
        self._loaders = self.query_inst_cfg("loaders", override=kwargs)
        # self._cache = self.query_inst_cfg("cache", override=kwargs)
        self._cache = True  # TOFIX

        self._extra_fields_enabled = self.query_inst_cfg(
            "extra_fields",
            override=kwargs,
            default=True,  # TOFIX, should be set to false by default
        )
        self._extra_fields = {}
        self._children_class = self.query_inst_cfg(
            "children_class", override=kwargs, default=NOT_SET
        )

        self._cast = self.query_inst_cfg("cast", override=kwargs, default=None)
        self._strict_cast = self.query_inst_cfg("strict_cast", override=kwargs)

        self._default = self.query_inst_cfg("default", override=kwargs, default=NOT_SET)
        if self._default is NOT_SET:
            self._default = self.query_parent_cfg(
                "default", as_subkey=True, default=NOT_SET
            )

        # print ("\n\n===== CREATE NEW CONFIG", self.key, self, value)
        child_values = self._default if self._value is NOT_SET else self._value
        self.set_values(child_values, lookup=True)

    def set_dyn_children(self, value):
        "Placeholder"

    # Generic API
    # ----------------------------

    # Field compat API
    @property
    def default(self):
        "Temporary property to access to self._default"
        return self._default

    @property
    def cast(self):
        "Temporary property to access to self._default"
        return self._cast

    # Field compatibility layer !
    # This basically respect default python behavior , when this is a children...
    def __get__(self, conf_instance, owner):
        # if conf_instance:
        #     return conf_instance.get_field_value(field=self)
        return self

    def __getitem__(self, value):
        return self.declared_fields[value].__get__(self, self.__class__)

    def __len__(self):
        return len(self.declared_fields)

    # def __repr__(self):
    #     return "{}(loaders=[{}])".format(
    #         self.__class__.__name__,
    #         ", ".join([str(loader) for loader in self._loaders]),
    #     )

    # def __str__(self):
    #     values = []
    #     for _, v in self:
    #         if v.default is NOT_SET and not v.help:
    #             help = "No default value provided"
    #         elif not v.help:
    #             help = "Default value is {}.".format(repr(v.default))
    #         else:
    #             help = v.help
    #         try:
    #             values.append(
    #                 "{}={} - {}".format(v.key, repr(getattr(self, v.key)), help)
    #             )
    #         except UnknownConfiguration:
    #             values.append("{}=NOT_SET - {}".format(v.key, help))
    #     return "\n".join(values)

    # Value management
    # ----------------------------

    def get_value(self, key, lvl=-1, **kwargs):
        """Get configuration value for a given key.

        Args:
            key: Configuration key to retrieve
            lvl: Recursion level for nested configurations
            **kwargs: Additional arguments passed to get_field_value

        Returns:
            Configuration value for the specified key
        """
        assert isinstance(key, str)
        return self.get_field_value(key, **kwargs)

    def reset(self):
        """Reset all loaders and clear cached values.

        This should be called when configuration values need to be reloaded.
        """
        for loader in self._loaders:
            loader.reset()
        self._cached_values = {}

    def get_field_value(self, key=None, field=None, default=UNSET_ARG, **kwargs):
        """Get value for a configuration field.

        Args:
            key: Configuration key name
            field: Configuration field object
            default: Default value if not found
            **kwargs: Additional arguments for child creation

        Returns:
            Configuration value for the specified field

        Raises:
            UndeclaredField: If field is not found and no default provided
        """

        # Parse input
        if field is None and key is None:
            assert False, "BUG here"

        if field is None:
            assert isinstance(key, (str, int))

            field = self.declared_fields.get(key, None)
            if field is None:
                if default is not UNSET_ARG:
                    return default
                raise exceptions.UndeclaredField(f"Configuration '{key}' not found")
            assert key == field.key, f"Got: {key} != {field.key}"

        if key is None:
            key = field.key

        # Check in cache
        if self._cache and key in self._cached_values:
            return self._cached_values[key]

        conf = self.create_child(key, field, **kwargs)
        assert isinstance(
            conf, (type(None), bool, int, str, Sequence, Mapping, ConfigurationDict)
        ), f"Got: {type(conf)}"

        if self._cache:
            self._cached_values[key] = conf
        return conf

    def get_values(self, lvl=-1, **kwargs):
        "Return all values of the container"

        if lvl == 0:
            return self

        out = {}
        for key, _ in self.declared_fields.items():
            val = self.get_field_value(key)
            if isinstance(val, Store):
                val = val.get_values(lvl=lvl - 1)

            out[key] = val

        return out

    # This should be split if field has children or not ...
    def create_child(self, key, field, value=NOT_SET, **kwargs):
        """
        :param item:    Name of the setting to lookup.
        :param default: Default value if none is provided. If left unset,
                        loading a self that fails to provide this value
                        will raise a UnknownConfiguration exception.
        :param cast:    Callable to cast variable with. Defaults to type of
                        default (if provided), identity if default is not
                        provided or raises TypeError if provided cast is not
                        callable.
        :param loaders: A list of loader instances in the order they should be
                        looked into. Defaults to `[Environment()]`
        """

        # General lookup policy
        #  - kwargs default override
        #  - current object defaults
        #      - Must be a dict, and the key must be present or NEXT
        #  - child
        #      - DEfault must be set
        #  - UNSET

        # DElegate logic to field methods
        result, meta = field.resolve_value(
            self,
            value=value,
        )

        # TOFIX: To be migrated into FieldConf
        default = meta.default
        value = meta.value

        # print ("DUMP CHILD CREATE META", self, key)
        # pprint (meta.__dict__)

        # If not container, return HERE
        if not isinstance(field, FieldConf):
            return result

        # Default children_class
        children_class = field.children_class
        if children_class is NOT_SET:
            children_class = self._children_class
            # children_class = getattr(field, "children_class", NOT_SET)

        assert (
            children_class
        ), f"Got: {type(children_class)}: {children_class} for {self}:{key}"

        out = children_class(
            key=key, value=value, default=default, parent=self, **kwargs
        )

        return out

    @property
    def declared_fields(self):
        "Return declared fields"
        out = {}
        if self._extra_fields:
            # Add extra fields
            out.update(self._extra_fields)

        # Always use explicit fields
        out.update(self._declared_values)
        return out

    def set_values(self, value, lookup=False):
        "Set a value"

        self.set_dyn_children(value)

        # Instanciate containers fields - Automatic
        for key, field in self.declared_fields.items():

            # Create child then
            val = NOT_SET
            if value and isinstance(value, dict):
                try:
                    val = value.get(key, NOT_SET)
                except AttributeError:
                    val = NOT_SET
            if value and isinstance(value, list):
                try:
                    val = value[key]
                except IndexError:
                    val = NOT_SET

            if field.is_container():
                logger.debug(
                    "Create child container %s.%s: %s=%s", self, key, field, val
                )
                # print ("AUTOMATIC CREATE CHILD CONTAINER", key, field, val)
                conf = self.create_child(key, field, value=val)
                assert isinstance(conf, (Store)), f"Got: {type(conf)}"
                self._cached_values[key] = conf
            else:

                result, _ = field.resolve_value(
                    self,
                    value=val,
                    lookup=lookup,
                )

                self._value = self._value or {}
                self._value[key] = result
                self._cached_values[key] = result


class DeclarativeValuesMetaclass(type):
    """
    Collect Value objects declared on the base classes
    """

    def __new__(mcs, class_name, bases, attrs):
        # Collect values from current class and all bases.
        values = OrderedDict()

        # Walk through the MRO and add values from base class.
        for base in reversed(bases):
            if hasattr(base, "_declared_values"):
                values.update(base._declared_values)

        for key, value in attrs.items():
            if isinstance(value, Field):
                if value.key and key != value.key:
                    raise AttributeError(
                        "Don't explicitly set keys when declaring values"
                    )
                value.key = key
                values.update({key: value})

        attrs["_declared_values"] = values

        return super(DeclarativeValuesMetaclass, mcs).__new__(
            mcs, class_name, bases, attrs
        )

    @classmethod
    def __prepare__(mcs, name, bases, **kwds):
        # Remember the order that values are defined.
        return OrderedDict()


class ConfigurationDict(Store, metaclass=DeclarativeValuesMetaclass):
    """Dictionary-based configuration container.

    Provides a dictionary interface to configuration values and supports
    dynamic field creation based on input values.
    """

    # meta__custom_field = "My VALUUUUuuueeeee"
    meta__loaders = [Environment()]
    meta__cache = True  # Yes by default ...
    meta__extra_fields = True
    meta__strict_cast = False

    # Optional fields
    # meta__default = NOT_SET # dict()
    # meta__extra_fields = NOT_SET # dict()

    def set_dyn_children(self, value):
        """Set up dynamic children based on input value.

        Creates fields dynamically for dictionary values that don't have
        corresponding declared fields.

        Args:
            value: Dictionary of configuration values
        """

        # Create children method
        # Check for predefined Fields
        # If additional_items == True
        # Check value
        # For each keys, check if a type exists, or field
        # Add to _extra_fields

        # For each children,
        # If class of Configuration, create child
        # If field, do noting

        # declared_fields = self.declared_fields
        children_class = self._children_class

        # Add extra fields
        child_values = value or dict()

        if isinstance(child_values, dict):

            # Look for new keys in value
            assert isinstance(
                child_values, dict
            ), f"Got {self}: {type(child_values)}: {child_values}"

            for key, _ in child_values.items():

                # Get best children_class
                field = None
                child_class = NOT_SET

                # Check if key have an existing field
                if key in self.declared_fields:
                    field = self.declared_fields[key]
                    # child_class = field.children_class
                    child_class = getattr(field, "children_class", NOT_SET)

                # Prevent unexpected childrens ...
                if not field and self._extra_fields_enabled is False:
                    msg = f"Undeclared key '{key}' for {self}, or enable extra_fields=True"
                    raise exceptions.UnknownExtraField(msg)

                if child_class is NOT_SET:
                    # Get children class form container
                    child_class = children_class

                if not field:
                    xtra_kwargs = {}
                    if not child_class:
                        # No children_class, then it's just a field
                        child_cls = Field
                    else:
                        child_cls = FieldConf
                        xtra_kwargs = dict(children_class=child_class)

                    logger.debug(
                        "Register dynamic field %s.%s: %s", self, key, child_cls
                    )

                    # Create dynamic field
                    field = child_cls(
                        key=key,
                        **xtra_kwargs,
                    )
                    self._extra_fields[key] = field

    def __iter__(self):
        yield from self.declared_fields.items()
        # yield from self._declared_values.items()


# ====================================
# Configuration Child (Dict)
# ====================================


class Configuration(ConfigurationDict):
    """Main configuration class supporting declarative field definitions.

    This class allows fields to be declared as class attributes and provides
    a clean interface for defining configuration schemas.
    """

    meta__extra_fields = False


# ====================================
# Configuration Container (List)
# ====================================


class ConfigurationList(Store):
    """List-based configuration container.

    Provides a list interface to configuration values and supports
    dynamic field creation for list elements.
    """

    # _declared_values = {}
    meta__loaders = [Environment()]
    meta__cache = True  # Yes by default ...
    meta__extra_fields = True
    meta__strict_cast = False

    def set_dyn_children(self, value):
        """Set up dynamic children based on input value.

        Creates fields dynamically for list elements.

        Args:
            value: List of configuration values
        """

        children_class = self._children_class
        child_values = value or []

        if isinstance(child_values, list):

            # Look for new keys in value
            assert isinstance(
                child_values, list
            ), f"Got {self}: {type(child_values)}: {child_values}"

            for key, _ in enumerate(child_values):

                # Get best children_class
                # field = None
                child_class = children_class

                xtra_kwargs = {}
                if not child_class:
                    # No children_class, then it's just a field
                    child_cls = Field
                else:
                    child_cls = FieldConf
                    xtra_kwargs = dict(children_class=child_class)

                # Create dynamic field
                field = child_cls(
                    key=key,
                    **xtra_kwargs,
                )
                self._extra_fields[key] = field

    def get_values(self, lvl=-1, **kwargs):
        """Get all configuration values as a list.

        Args:
            lvl: Recursion level for nested configurations
            **kwargs: Additional arguments passed to parent method

        Returns:
            List of configuration values
        """
        out = super().get_values(lvl=lvl, **kwargs)

        if isinstance(out, Mapping):
            out = list(out.values())
        return out

    def __getitem__(self, value):
        value = int(value)

        return super().__getitem__(value)



"Common code"

# pylint: skip-file


import json
import logging
import os
import re

import yaml

log = logging.getLogger(__name__)


# Setup YAML object
# yaml = ruamel.yaml.YAML()
# yaml.version = (1, 1)
# yaml.default_flow_style = False
# # yaml.indent(mapping=3, sequence=2, offset=0)
# yaml.allow_duplicate_keys = True
# yaml.explicit_start = True


# pylint: disable=redefined-builtin
def truncate(data, max=72, txt=" ..."):
    "Truncate a text to max lenght and replace by txt"
    data = str(data)
    if max < 0:
        return data
    if len(data) > max:
        return data[: max + len(txt)] + txt
    return data


# TODO: Add tests on this one
def to_domain(string, sep=".", alt="-"):
    "Transform any string to valid domain name"

    domain = string.split(sep)
    result = []
    for part in domain:
        part = re.sub("[^a-zA-Z0-9]", alt, part)
        part.strip(alt)
        result.append(part)

    return ".".join(result)


# TODO: Add tests on this one
def first(array):
    "Return the first element of a list or None"
    # return next(iter(array))
    array = list(array) or []
    result = None
    if len(array) > 0:
        result = array[0]
    return result


# TODO: add tests
def from_yaml(string):
    "Transform YAML string to python dict"
    return yaml.safe_load(string)


# TODO: add tests
def to_yaml(obj, headers=False):
    "Transform obj to YAML"
    options = {}
    string_stream = StringIO()

    if isinstance(obj, str):
        obj = json.loads(obj)

    yaml.dump(obj, string_stream, **options)
    output_str = string_stream.getvalue()
    string_stream.close()
    if not headers:
        output_str = output_str.split("\n", 2)[2]
    return output_str


# TODO: add tests
def to_json(obj, nice=True):
    "Transform JSON string to python dict"
    if nice:
        return json.dumps(obj, indent=2)
    return json.dumps(obj)


# TODO: add tests
def from_json(string):
    "Transform JSON string to python dict"
    return json.loads(string)


# TODO: add tests
def to_dict(obj):
    """Transform JSON obj/string to python dict

    Useful to transofmr nested dicts as well"""
    if not isinstance(obj, str):
        obj = json.dumps(obj)
    return json.loads(obj)


def duplicates(_list):
    """Check if given list contains duplicates"""
    known = set()
    dup = set()
    for item in _list:
        if item in known:
            dup.add(item)
        else:
            known.add(item)

    if len(dup) > 0:
        return list(dup)
    return []


def read_file(file):
    "Read file content"
    with open(file, encoding="utf-8") as _file:
        return "".join(_file.readlines())


def write_file(file, content):
    "Write content to file"

    file_folder = os.path.dirname(file)
    if not os.path.exists(file_folder):
        os.makedirs(file_folder)

    with open(file, "w", encoding="utf-8") as _file:
        _file.write(content)


def flatten(array):
    "Flatten any arrays nested arrays"
    if array == []:
        return array
    if isinstance(array[0], list):
        return flatten(array[0]) + flatten(array[1:])
    return array[:1] + flatten(array[1:])


def list_parent_dirs(path):
    """
    Return a list of the parents paths
    path treated as strings, must be absolute path
    """
    result = [path]
    val = path
    while val and val != os.sep:
        val = os.path.split(val)[0]
        result.append(val)
    return result


def find_file_up(names, paths):
    """
    Find every files names in names list in
    every listed paths
    """
    assert isinstance(names, list), f"Names must be array, not: {type(names)}"
    assert isinstance(paths, list), f"Paths must be array, not: {type(names)}"

    result = []
    for path in paths:
        for name in names:
            file_path = os.path.join(path, name)
            if os.access(file_path, os.R_OK):
                result.append(file_path)

    return result


def filter_existing_files(root_path, candidates):
    """Return only existing files"""
    result = [
        os.path.join(root_path, cand)
        for cand in candidates
        if os.path.isfile(os.path.join(root_path, cand))
    ]
    return list(set(result))


def ensure_dir_exists(path):
    """Ensure directories exist for a given path"""
    if not os.path.isdir(path):
        log.info(f"Create directory: {path}")
        os.makedirs(path)
        return True
    return False


def ensure_parent_dir_exists(path):
    """Ensure parent directories exist for a given path"""
    parent = os.path.dirname(os.path.normpath(path))
    return ensure_dir_exists(parent)



"""Path anchoring and manipulation library for flexible path resolution.

This module provides classes for handling paths with anchor points, allowing for flexible
path resolution relative to different base directories. It's particularly useful for
configuration management and file organization where paths need to be resolved relative
to different root directories.

Main Components:
    - PathAnchor: Base class for handling directory paths with anchor points
    - FileAnchor: Extension of PathAnchor for handling file paths specifically

Key Features:
    - Path resolution relative to anchor points
    - Support for both absolute and relative path modes
    - Chain of anchor points for complex path hierarchies
    - Clean path normalization
    - Flexible path representation

Examples:
    Basic path anchoring:

    >>> project_dir = "/fake/prj"
    >>> root = PathAnchor(project_dir)
    >>> conf = PathAnchor("../../common_conf", anchor=root, mode="abs")
    >>> inventory = PathAnchor("inventory/", anchor=conf, mode="rel")
    >>> 
    >>> # Get paths in different modes
    >>> conf.get_dir()  # Returns absolute path
    '/fake/common_conf'
    >>> inventory.get_dir()  # Returns relative path
    '../../inventory'

    File handling:

        >>> root = PathAnchor("/fake/prj")
        >>> config_file = FileAnchor("subconf/myfile.yml", anchor=root)
        >>> config_file.get_path()  # Full path
        '/fake/prj/subconf/myfile.yml'
        >>> config_file.get_file()  # Just filename
        'myfile.yml'
        >>> config_file.get_dir()   # Just directory
        '/fake/prj/subconf'

    Complex path resolution:

        >>> project = PathAnchor("/fake_root/project")
        >>> path = PathAnchor("subdir2/../../subdir2/file", anchor=project)
        >>> path.get_dir(clean=True)  # Normalizes the path
        '/fake_root/subdir2/file'
"""

import os

# from pprint import pprint
from typing import List, Optional


class PathAnchor:
    """A class representing a path with an optional anchor point and display mode.

    This class handles path operations with support for anchored paths, allowing paths
    to be relative to a specified anchor point. It also supports different display modes
    (absolute or relative paths).

    Attributes:
        path_anchor: The anchor point for relative paths
        path_dir: The directory path
        path_mode: The display mode ('abs' or 'rel')
    """

    def __init__(
        self,
        path: str,
        mode: Optional[str] = None,
        anchor: Optional["PathAnchor"] = None,
    ):
        """Initialize a PathAnchor object.

        Args:
            path (str): The path to use
            mode (Optional[str], optional): Display mode for path rendering.
                Can be 'abs' for absolute paths or 'rel' for relative paths.
                If None, the mode remains unchanged. Defaults to None.
            anchor (Optional[PathAnchor], optional): Another PathAnchor object to use as
                reference point.
                Defaults to None.
        """
        self.path_anchor = anchor
        self.path_dir = path
        self.path_mode = mode

    def __repr__(self) -> str:
        """Return a string representation of the PathAnchor object.

        Returns:
            str: A string representation showing the path, anchor (if present), and mode (if set)
        """
        name = self.__class__.__name__
        ret = f"<{name} {self.get_path()}"

        # Change if anchored
        anchor = self.path_anchor
        if anchor:
            ret = f"<{name} [{anchor.get_dir()}]{self.get_path()}"

        # Add suffix
        suffix = ">"
        if self.path_mode:
            suffix = f" (mode={self.path_mode})>"
        return ret + suffix

    def get_mode(self, lvl: int = 0) -> Optional[str]:
        """Get the effective display mode for this path.

        If this object has no mode set but has an anchor, it will recursively
        check the anchor's mode.

        Args:
            lvl (int, optional): Current recursion level. Defaults to 0.

        Returns:
            Optional[str]: The effective mode ('abs' or 'rel') or None if no mode is set
        """
        if isinstance(self.path_mode, str):
            return self.path_mode

        if self.path_anchor:
            lvl += 1
            return self.path_anchor.get_mode(lvl=lvl)
        return None

    def get_parents(self, itself: bool = False) -> List:
        """Get a list of all parent anchors in the anchor chain.

        Args:
            itself (bool, optional): Include the current object in the result if True.
                Defaults to False.

        Returns:
            list[PathAnchor]: List of parent PathAnchor objects, ordered from current to root
        """
        ret = []
        ret.append(self)

        if self.path_anchor:
            tmp = self.path_anchor.get_parents(itself=True)
            ret.extend(tmp)

        if not itself:
            ret = ret[1:]
        return ret

    def get_dir(
        self,
        mode: Optional[str] = None,
        clean: Optional[bool] = False,
        start: Optional[str] = None,
        anchor: Optional["PathAnchor"] = None,
    ) -> str:
        """Get the directory path according to specified parameters.

        Args:
            mode (Optional[str], optional): Output path format.
                Can be 'abs' for absolute path or 'rel' for relative path.
                If None, uses the object's mode setting. Defaults to None.
            clean (Optional[bool], optional): If True, normalizes the path by resolving
                '..' and '.' components. Defaults to False.
            start (Optional[str], optional): Base directory for relative path calculation
                when mode is 'rel'. Defaults to current working directory.
            anchor (Optional[PathAnchor], optional): Override the anchor point for this
                operation. Defaults to None.

        Returns:
            str: The processed directory path

        Raises:
            AssertionError: If an invalid mode is specified
        """
        ret = None
        mode = mode or self.get_mode()
        start = start or os.getcwd()

        # Resolve name
        if os.path.isabs(self.path_dir):
            ret = self.path_dir
        else:
            anchor = anchor or self.path_anchor
            if anchor:
                ret = os.path.join(anchor.path_dir, self.path_dir)
            else:
                ret = self.path_dir

        # Clean
        if clean:
            ret = os.path.normpath(ret)

        # Ensure output format
        if mode == "rel":
            if os.path.isabs(ret):
                ret = os.path.relpath(ret, start=start)
        elif mode == "abs":
            if not os.path.isabs(ret):
                ret = os.path.abspath(ret)
        elif mode is None:
            pass
        else:
            raise ValueError(f"Invalid mode: {mode}")

        return ret

    def get_path(self, **kwargs) -> str:
        """Get the path using the same parameters as get_dir.

        Returns:
            str: The processed path
        """
        return self.get_dir(**kwargs)


class FileAnchor(PathAnchor):
    """A class representing a file path with optional anchor point and display mode.

    This class extends PathAnchor to handle file paths specifically, maintaining
    separate tracking of directory and filename components.

    Attributes:
        path_dir: The directory component of the path
        path_file: The filename component of the path
        path_anchor: The anchor point for relative paths
        path_mode: The display mode ('abs' or 'rel')
    """

    # pylint: disable=too-many-arguments, too-many-positional-arguments
    def __init__(
        self,
        path: str,
        directory: Optional[str] = None,
        filename: Optional[str] = None,
        mode: Optional[str] = None,
        anchor: Optional["PathAnchor"] = None,
    ):
        """Initialize a FileAnchor object.

        The path can be specified either as a full path or as separate directory
        and file components.

        Args:
            path (str): The complete file path
            directory (Optional[str], optional): The directory component. If provided with filename,
                path parameter is ignored. Defaults to None.
            filename (Optional[str], optional): The filename component. Defaults to None.
            mode (Optional[str], optional): Display mode for path rendering.
                Can be 'abs' for absolute paths or 'rel' for relative paths.
                If None, the mode remains unchanged. Defaults to None.
            anchor (Optional[PathAnchor], optional): Another PathAnchor object to use
                as reference point. Defaults to None.

        Raises:
            AssertionError: If the provided path components don't match the expected format
        """
        if directory and filename:
            # Ignore path
            path = os.path.join(directory, filename)
        elif path and filename:
            path = os.path.join(path, filename)
            directory = path
        elif path and directory:
            assert path.startswith(directory)
            filename = path[len(directory) :]
        elif path:
            directory, filename = os.path.split(path)

        assert path == os.path.join(
            directory, filename
        ), f"Got: {path} == {os.path.join(directory, filename)}"

        super().__init__(directory, mode=mode, anchor=anchor)
        self.path_file = filename

    def get_path(self, **kwargs) -> str:
        """Get the complete file path.

        This method combines the directory path from get_dir() with the filename.
        All kwargs are passed to get_dir().

        Returns:
            str: The complete file path (directory + filename)
        """
        return os.path.join(self.get_dir(**kwargs), self.get_file())

    def get_file(self) -> str:
        """Get the filename component of the path.

        Returns:
            str: The filename without directory path
        """
        return self.path_file



"XDG Support"

# pylint: disable=too-few-public-methods, too-many-nested-blocks, too-many-branches

import logging
import os
from pathlib import Path

# pylint: disable=unused-import
from pprint import pprint

from superconf.casts import AsBest, AsList, AsString
from superconf.common import (
    from_json,
    from_yaml,
    read_file,
    to_json,
    to_yaml,
    write_file,
)
from superconf.configuration import ConfigurationDict
from superconf.fields import Field

logger = logging.getLogger(__name__)


# ====================================
# Exceptions and casts
# ====================================


class XDGException(Exception):
    "XDG error"


class CastEnvColonValues(AsList):
    "Cast env value as list with colon seprated values"

    def __init__(self, **kwargs):

        self.as_best = AsBest()
        super().__init__(**kwargs)

    def __call__(self, value):
        # print ("CALL CastEnvColonValues", self, value )
        # out = traceback.print_stack()
        # print ("EXECUTED", out)

        out = super().__call__(value)
        ret = []
        for x in out:
            x = os.path.expandvars(x)
            x = self.as_best(x)
            ret.append(x)

        return ret


# ====================================
# XDG Configuration
# ====================================

# Create cast instances
cast_env = AsBest()  # Try fetch best type
cast_env_colon = AsList(delimiter=":")  # We force lists with delimiter


class UserConfig(ConfigurationDict):
    "User var support"

    meta__env_parents = False  # Not parent keys
    meta__env_name = ""  # No prefix for environment vars

    HOME = Field(default="/home/user", cast=cast_env)
    USER = Field(default="user", cast=cast_env)
    UID = Field(default="1000", cast=cast_env)
    LANG = Field(default="en_US.utf8", cast=cast_env)

    SHELL = Field(default="/bin/bash", cast=cast_env)
    SHLVL = Field(default="/bin/bash", cast=cast_env)
    PWD = Field(default="/", cast=cast_env)
    OLDPWD = Field(default="/", cast=cast_env)
    PATH = Field(default="/", cast=cast_env_colon)
    DISPLAY = Field(default="", cast=cast_env)

    WEIRD = Field(cast=cast_env)


class XDGConfig(UserConfig):
    "Base class with XDG var support"

    meta__xdg_file_fmt = [
        "yml",
        "yaml",
        "json",
        "toml",
        # "ini",
    ]

    XDG_CONFIG_HOME = Field(default="$HOME/.config", cast=cast_env)  # Like: /etc
    XDG_DATA_HOME = Field(
        default="$HOME/.local/share", cast=cast_env
    )  # Like: /usr/share
    XDG_CACHE_HOME = Field(default="$HOME/.cache", cast=cast_env)  # Like: /var/cache
    XDG_STATE_HOME = Field(
        default="$HOME/.local/state", cast=cast_env
    )  # Like: /var/lib

    XDG_RUNTIME_DIR = Field(
        default="/run/user/$UID", cast=cast_env
    )  # Like: /run/user/$UID for pam_systemd

    XDG_CONFIG_DIRS = Field(default="/etc/xdg", cast=cast_env_colon)
    XDG_DATA_DIRS = Field(default="/usr/local/share:/usr/share", cast=cast_env_colon)

    # Legacy commands
    def get_config_file(self, name=None):
        """Get config file path."""
        return self.get_file("XDG_CONFIG_HOME", name)

    def get_config_dir(self, name=None):
        """Get config directory path."""
        return self.get_dir("XDG_CONFIG_HOME", name)

    # Internal API
    def _parse_path(self, values, name=None, prefix=None, extensions=None):
        """Parse and expand path."""

        found = False
        extensions = extensions or []
        ret = []
        for path in values:

            path = os.path.expandvars(path)
            if isinstance(prefix, str):
                path = os.path.join(path, prefix)
            path = path.rstrip("/")

            if extensions:
                # File lookup only

                if name:

                    # Validate name doesn't already end with any extension
                    name, file_extension = os.path.splitext(name)
                    if file_extension:
                        file_extension = file_extension.lstrip(".")
                        if file_extension not in extensions:
                            name = f"{name}.{file_extension}"
                        else:
                            extensions = [file_extension]

                    default_path = os.path.join(path, f"{name}.{extensions[0]}")
                    found = False
                    for ext in extensions:
                        if found is False:
                            fname = f"{name}.{ext}"
                            fpath = os.path.join(path, fname)
                            if os.path.isfile(fpath):
                                path = fpath
                                found = True
                else:
                    default_path = f"{path}.{extensions[0]}"
                    found = False
                    for ext in extensions:
                        if found is False:
                            fpath = f"{path}.{ext}"
                            if os.path.isfile(fpath):
                                path = fpath
                                found = True

                path = path if found else default_path
            else:
                # Only for directory lookup
                if name:
                    path = os.path.join(path, f"{name}")

                else:
                    path = f"{path}"

            path = Path(path)
            ret.append(path)

        # print ("MATCHED", found, ret)
        return ret

    def read_file(self, item, name=None, missing_ok=True):
        "Read a file content"

        # Fetch best file
        files = self.get_file(item, name=name)
        if not isinstance(files, list):
            files = [files]

        # Return error if no files
        if not files:
            if name:
                raise XDGException(f"Could not find any file in {item} called {name}")
            raise XDGException(f"Could not find any file in {item}")

        # Return first file
        out = None
        found = False
        for file in files:
            file = str(file)

            if not os.path.isfile(file):
                continue

            found = True
            logger.info("Read file %s", file)

            fcontent = read_file(file)
            if file.endswith("yaml") or file.endswith("yml"):
                out = from_yaml(fcontent)
            elif file.endswith("json"):
                out = from_json(fcontent)
            elif file.endswith("toml"):
                raise NotImplementedError("Toml support not implemented yet")
            elif file.endswith("ini"):
                raise NotImplementedError("Ini support is not implemented yet")
            else:
                raise NotImplementedError(f"Format not supported: {file}")

            break

        if not found:
            if not missing_ok:
                raise XDGException(f"Could not find any file in {str(files[0])}")

        return out

    def write_file(self, item, data, name=None):
        "Write content to file"

        # Fetch best file
        files = self.get_file(item, name=name)
        if not isinstance(files, list):
            files = [files]

        # Return error if no files
        if not files:
            if name:
                raise XDGException(f"Could not find any file in {item} called {name}")
            raise XDGException(f"Could not find any file in {item}")

        # Write only on the first file
        files = [files[0]]
        for file in files:
            file = str(file)

            # fcontent = read_file(file)
            if file.endswith("yaml") or file.endswith("yml"):
                out = to_yaml(data)
            elif file.endswith("json"):
                out = to_json(data)
            elif file.endswith("toml"):
                raise NotImplementedError("Toml support not implemented yet")
            elif file.endswith("ini"):
                raise NotImplementedError("Ini support is not implemented yet")
            else:
                raise NotImplementedError(f"Format not supported: {file}")

            logger.info("Write data in file %s", file)
            write_file(file, out)
            break

    def get_file(self, item, name=None):
        """Get file path for given XDG item."""

        # Fetch current value
        if not hasattr(self, item.upper()):
            choices = list(self.get_values().keys())
            raise XDGException(f"Unknown item: {item}, choose one of: {choices}")
        values = self[item]

        # Convert to list
        return_one = True
        if isinstance(values, list):
            return_one = False
        else:
            values = [values]
        assert isinstance(values, list), f"Got: {values}"

        # Prepare extensions
        extensions = list(self.meta__xdg_file_fmt)
        if extensions:  # Only append first extension if list is not empty
            extensions.append(extensions[0])

        # Parse paths
        app_name = self.query_cfg("app_name", default=None)
        if not isinstance(app_name, str):
            root = self.get_hierarchy()[-1]
            app_name = root.__class__.__name__
        assert app_name

        # print ("=== PARSE PATH", values, name, app_name, extensions)
        ret = self._parse_path(
            values, name=name, prefix=app_name, extensions=extensions
        )

        if return_one is True:
            return ret[0] if len(ret) > 0 else None

        return ret

    def get_dir(self, item, name=None):
        """Get directory path for given XDG item."""

        # Fetch current value
        if not hasattr(self, item.upper()):
            choices = list(self.get_values().keys())
            raise XDGException(f"Unknown item: {item}, choose one of: {choices}")
        values = self[item]

        # Convert to list
        return_one = True
        if isinstance(values, list):
            return_one = False
        else:
            values = [values]
        assert isinstance(values, list), f"Got: {values}"

        # Prepare extensions
        extensions = []

        # Parse paths
        app_name = self.query_cfg("app_name", default=None)
        if not app_name or not isinstance(app_name, str):
            root = self.get_hierarchy()[-1]
            app_name = root.__class__.__name__

        ret = self._parse_path(
            values, name=name, prefix=app_name, extensions=extensions
        )

        if return_one is True:
            return ret[0] if len(ret) > 0 else None

        return ret

    # def _process_vars(self, item, name=None):
    #     """Process XDG variables into paths."""
    #     paths = self._resolve_var(item)
    #     out = []
    #     for path in paths:
    #         parsed = self._parse_path(path)
    #         if name:
    #             parsed = parsed / name
    #         out.append(parsed)
    #     return out

    # def _resolve_var(self, item):
    #     """Resolve XDG variable to list of paths."""
    #     if not hasattr(self, item):
    #         choices = list(self.get_values().keys())
    #         raise XDGException(f"Unknown item: {item}, choose one of: {choices}")

    #     value = self[item]
    #     if isinstance(value, str):
    #         value = [value]
    #     assert isinstance(value, list), f"Got: {value}"
    #     return value

    # def _parse_path(self, path):
    #     """Parse and expand path."""
    #     app_name = self.query_cfg("app_name")
    #     out = os.path.join(path, app_name)
    #     out = os.path.expandvars(out)
    #     return Path(out)

    # def find_first_file(self, pdirs):
    #     "Return configuration first existing configfile"

    #     extensions = self.query_cfg("xdg_file_fmt")
    #     extensions.append(extensions[0])

    #     # Try each file
    #     fname = None
    #     for pdir in pdirs:
    #         pdir = pdir.absolute()
    #         for ext in extensions:
    #             fname = f"{pdir}.{ext}"
    #             if os.path.isfile(fname):
    #                 break

    #     return Path(fname)


class ExtraConfig(XDGConfig):
    "Extra var support"

    GIT_AUTHOR_EMAIL = Field()
    GIT_AUTHOR_NAME = Field()
    GIT_COMMITTER_EMAIL = Field()
    GIT_COMMITTER_NAME = Field()

    TERM = Field()
    SSH_AGENT_PID = Field()
    SSH_AUTH_SOCK = Field()
    XDG_SEAT = Field()
    XDG_SEAT_PATH = Field()
    XDG_SESSION_CLASS = Field()
    XDG_SESSION_DESKTOP = Field()
    XDG_SESSION_ID = Field()
    XDG_SESSION_PATH = Field()
    XDG_SESSION_TYPE = Field()
    VIRTUAL_ENV = Field()
    LOGNAME = Field()
    GNUPGHOME = Field()
    # GITEA_LOGIN = Field()
    # GITEA_SERVER_URL = Field()
    # GH_REPO = Field()
    # GH_TOKEN = Field()
    # DIRENV_DIR = Field()
    # DIRENV_FILE = Field()
    DESKTOP_SESSION = Field()
    DBUS_SESSION_BUS_ADDRESS = Field()
